@@ -40,12 +40,16 @@ struct connman_service {
 	enum connman_service_mode mode;
 	enum connman_service_security security;
 	enum connman_service_state state;
+	enum connman_service_error error;
 	connman_uint8_t strength;
 	connman_bool_t favorite;
+	connman_bool_t hidden;
+	GTimeVal modified;
 	unsigned int order;
 	char *name;
 	char *passphrase;
 	char *profile;
+	struct connman_ipconfig *ipconfig;
 	struct connman_device *device;
 	struct connman_network *network;
 	DBusMessage *pending;
@@ -71,6 +75,34 @@ void __connman_service_list(DBusMessageIter *iter)
 	g_sequence_foreach(service_list, append_path, iter);
 }
 
+struct find_data {
+	const char *path;
+	struct connman_service *service;
+};
+
+static void compare_path(gpointer value, gpointer user_data)
+{
+	struct connman_service *service = value;
+	struct find_data *data = user_data;
+
+	if (data->service != NULL)
+		return;
+
+	if (g_strcmp0(service->path, data->path) == 0)
+		data->service = service;
+}
+
+static struct connman_service *find_service(const char *path)
+{
+	struct find_data data = { .path = path, .service = NULL };
+
+	DBG("path %s", path);
+
+	g_sequence_foreach(service_list, compare_path, &data);
+
+	return data.service;
+}
+
 static const char *type2string(enum connman_service_type type)
 {
 	switch (type) {
@@ -82,6 +114,10 @@ static const char *type2string(enum connman_service_type type)
 		return "wifi";
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 		return "wimax";
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+		return "bluetooth";
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		return "cellular";
 	}
 
 	return NULL;
@@ -112,8 +148,8 @@ static const char *security2string(enum connman_service_security security)
 		return "wep";
 	case CONNMAN_SERVICE_SECURITY_WPA:
 		return "wpa";
-	case CONNMAN_SERVICE_SECURITY_WPA2:
-		return "wpa2";
+	case CONNMAN_SERVICE_SECURITY_RSN:
+		return "rsn";
 	}
 
 	return NULL;
@@ -143,6 +179,26 @@ static const char *state2string(enum connman_service_state state)
 	return NULL;
 }
 
+static const char *error2string(enum connman_service_error error)
+{
+	switch (error) {
+	case CONNMAN_SERVICE_ERROR_UNKNOWN:
+		break;
+	case CONNMAN_SERVICE_ERROR_DHCP_FAILED:
+		return "dhcp-failed";
+	}
+
+	return NULL;
+}
+
+static enum connman_service_error string2error(const char *error)
+{
+	if (g_strcmp0(error, "dhcp-failed") == 0)
+		return CONNMAN_SERVICE_ERROR_DHCP_FAILED;
+
+	return CONNMAN_SERVICE_ERROR_UNKNOWN;
+}
+
 static void state_changed(struct connman_service *service)
 {
 	DBusMessage *signal;
@@ -173,6 +229,36 @@ static void state_changed(struct connman_service *service)
 	g_dbus_send_message(connection, signal);
 }
 
+static void strength_changed(struct connman_service *service)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	const char *key = "Strength";
+
+	if (service->path == NULL)
+		return;
+
+	if (service->strength == 0)
+		return;
+
+	signal = dbus_message_new_signal(service->path,
+				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_BYTE_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BYTE,
+							&service->strength);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+}
+
 static DBusMessage *get_properties(DBusConnection *conn,
 					DBusMessage *msg, void *user_data)
 {
@@ -214,6 +300,11 @@ static DBusMessage *get_properties(DBusConnection *conn,
 		connman_dbus_dict_append_variant(&dict, "State",
 						DBUS_TYPE_STRING, &str);
 
+	str = error2string(service->error);
+	if (str != NULL)
+		connman_dbus_dict_append_variant(&dict, "Error",
+						DBUS_TYPE_STRING, &str);
+
 	if (service->strength > 0)
 		connman_dbus_dict_append_variant(&dict, "Strength",
 					DBUS_TYPE_BYTE, &service->strength);
@@ -231,6 +322,8 @@ static DBusMessage *get_properties(DBusConnection *conn,
 		connman_dbus_dict_append_variant(&dict, "Passphrase",
 				DBUS_TYPE_STRING, &service->passphrase);
 
+	__connman_ipconfig_append_ipv4(service->ipconfig, &dict, "IPv4.");
+
 	dbus_message_iter_close_container(&array, &dict);
 
 	return reply;
@@ -279,7 +372,43 @@ static DBusMessage *set_property(DBusConnection *conn,
 				"WiFi.Passphrase", service->passphrase);
 
 		__connman_storage_save_service(service);
-	}
+	} else if (g_str_has_prefix(name, "IPv4.") == TRUE) {
+		int err;
+
+		err = __connman_ipconfig_set_ipv4(service->ipconfig,
+							name + 5, &value);
+		if (err < 0)
+			return __connman_error_failed(msg, -err);
+	} else
+		return __connman_error_invalid_property(msg);
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+}
+
+static DBusMessage *clear_property(DBusConnection *conn,
+					DBusMessage *msg, void *user_data)
+{
+	struct connman_service *service = user_data;
+	const char *name;
+
+	DBG("service %p", service);
+
+	dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &name,
+							DBUS_TYPE_INVALID);
+
+	if (__connman_security_check_privilege(msg,
+					CONNMAN_SECURITY_PRIVILEGE_MODIFY) < 0)
+		return __connman_error_permission_denied(msg);
+
+	if (g_str_equal(name, "Error") == TRUE) {
+		service->state = CONNMAN_SERVICE_STATE_IDLE;
+		service->error = CONNMAN_SERVICE_ERROR_UNKNOWN;
+		state_changed(service);
+
+		g_get_current_time(&service->modified);
+		__connman_storage_save_service(service);
+	} else
+		return __connman_error_invalid_property(msg);
 
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
@@ -292,6 +421,9 @@ static gboolean connect_timeout(gpointer user_data)
 
 	service->timeout = 0;
 
+	if (service->network != NULL)
+		__connman_network_disconnect(service->network);
+
 	if (service->pending != NULL) {
 		DBusMessage *reply;
 
@@ -301,6 +433,9 @@ static gboolean connect_timeout(gpointer user_data)
 
 		dbus_message_unref(service->pending);
 		service->pending = NULL;
+
+		__connman_service_indicate_state(service,
+					CONNMAN_SERVICE_STATE_FAILURE);
 	}
 
 	return FALSE;
@@ -322,19 +457,24 @@ static DBusMessage *connect_service(DBusConnection *conn,
 	if (service->network != NULL) {
 		int err;
 
+		if (service->hidden == TRUE)
+			return __connman_error_invalid_service(msg);
+
 		connman_network_set_string(service->network,
 				"WiFi.Passphrase", service->passphrase);
 
-		err = connman_network_connect(service->network);
-		if (err < 0 && err != -EINPROGRESS)
-			return __connman_error_failed(msg, -err);
+		err = __connman_network_connect(service->network);
+		if (err < 0) {
+			if (err != -EINPROGRESS)
+				return __connman_error_failed(msg, -err);
 
-		service->pending = dbus_message_ref(msg);
+			service->pending = dbus_message_ref(msg);
 
-		service->timeout = g_timeout_add_seconds(45,
+			service->timeout = g_timeout_add_seconds(45,
 						connect_timeout, service);
 
-		return NULL;
+			return NULL;
+		}
 	} else if (service->device != NULL) {
 		if (service->favorite == FALSE)
 			return __connman_error_no_carrier(msg);
@@ -347,9 +487,10 @@ static DBusMessage *connect_service(DBusConnection *conn,
 						connect_timeout, service);
 
 		return NULL;
-	}
+	} else
+		return __connman_error_not_supported(msg);
 
-	return __connman_error_not_supported(msg);
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
 static DBusMessage *disconnect_service(DBusConnection *conn,
@@ -378,8 +519,6 @@ static DBusMessage *disconnect_service(DBusConnection *conn,
 		err = __connman_network_disconnect(service->network);
 		if (err < 0 && err != -EINPROGRESS)
 			return __connman_error_failed(msg, -err);
-
-		return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 	} else if (service->device != NULL) {
 		int err;
 
@@ -389,11 +528,10 @@ static DBusMessage *disconnect_service(DBusConnection *conn,
 		err = __connman_device_disconnect(service->device);
 		if (err < 0)
 			return __connman_error_failed(msg, -err);
+	} else
+		return __connman_error_not_supported(msg);
 
-		return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
-	}
-
-	return __connman_error_not_supported(msg);
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
 static DBusMessage *remove_service(DBusConnection *conn,
@@ -406,13 +544,11 @@ static DBusMessage *remove_service(DBusConnection *conn,
 	if (service->type == CONNMAN_SERVICE_TYPE_ETHERNET)
 		return __connman_error_not_supported(msg);
 
-	if (service->network != NULL) {
-		int err;
+	if (service->favorite == FALSE)
+		return __connman_error_not_supported(msg);
 
-		err = __connman_network_disconnect(service->network);
-		if (err < 0 && err != -EINPROGRESS)
-			return __connman_error_failed(msg, -err);
-	}
+	if (service->network != NULL)
+		__connman_network_disconnect(service->network);
 
 	connman_service_set_favorite(service, FALSE);
 	__connman_storage_save_service(service);
@@ -424,12 +560,37 @@ static DBusMessage *move_before(DBusConnection *conn,
 					DBusMessage *msg, void *user_data)
 {
 	struct connman_service *service = user_data;
+	struct connman_service *target;
+	const char *path;
+	GSequenceIter *src, *dst;
 
 	DBG("service %p", service);
 
+	dbus_message_get_args(msg, NULL, DBUS_TYPE_OBJECT_PATH, &path,
+							DBUS_TYPE_INVALID);
+
 	if (service->favorite == FALSE)
 		return __connman_error_not_supported(msg);
 
+	target = find_service(path);
+	if (target == NULL || target->favorite == FALSE || target == service)
+		return __connman_error_invalid_service(msg);
+
+	DBG("target %s", target->identifier);
+
+	g_get_current_time(&service->modified);
+	__connman_storage_save_service(service);
+
+	src = g_hash_table_lookup(service_hash, service->identifier);
+	dst = g_hash_table_lookup(service_hash, target->identifier);
+
+#if 0
+	g_sequence_move(src, dst);
+
+	__connman_profile_changed();
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+#endif
 	return __connman_error_not_implemented(msg);
 }
 
@@ -437,18 +598,33 @@ static DBusMessage *move_after(DBusConnection *conn,
 					DBusMessage *msg, void *user_data)
 {
 	struct connman_service *service = user_data;
+	struct connman_service *target;
+	const char *path;
 
 	DBG("service %p", service);
 
+	dbus_message_get_args(msg, NULL, DBUS_TYPE_OBJECT_PATH, &path,
+							DBUS_TYPE_INVALID);
+
 	if (service->favorite == FALSE)
 		return __connman_error_not_supported(msg);
 
+	target = find_service(path);
+	if (target == NULL || target->favorite == FALSE || target == service)
+		return __connman_error_invalid_service(msg);
+
+	DBG("target %s", target->identifier);
+
+	g_get_current_time(&service->modified);
+	__connman_storage_save_service(service);
+
 	return __connman_error_not_implemented(msg);
 }
 
 static GDBusMethodTable service_methods[] = {
 	{ "GetProperties", "",   "a{sv}", get_properties     },
 	{ "SetProperty",   "sv", "",      set_property       },
+	{ "ClearProperty", "s",  "",      clear_property     },
 	{ "Connect",       "",   "",      connect_service,
 						G_DBUS_METHOD_FLAG_ASYNC },
 	{ "Disconnect",    "",   "",      disconnect_service },
@@ -472,8 +648,10 @@ static void service_free(gpointer user_data)
 
 	g_hash_table_remove(service_hash, service->identifier);
 
-	if (service->timeout > 0)
+	if (service->timeout > 0) {
 		g_source_remove(service->timeout);
+		service->timeout = 0;
+	}
 
 	if (service->pending != NULL) {
 		dbus_message_unref(service->pending);
@@ -493,6 +671,8 @@ static void service_free(gpointer user_data)
 	if (service->network != NULL)
 		connman_network_unref(service->network);
 
+	connman_ipconfig_unref(service->ipconfig);
+
 	g_free(service->profile);
 	g_free(service->name);
 	g_free(service->passphrase);
@@ -533,6 +713,7 @@ static void __connman_service_initialize(struct connman_service *service)
 	service->state    = CONNMAN_SERVICE_STATE_UNKNOWN;
 
 	service->favorite = FALSE;
+	service->hidden = FALSE;
 
 	service->order = 0;
 }
@@ -556,6 +737,15 @@ struct connman_service *connman_service_create(void)
 
 	__connman_service_initialize(service);
 
+	service->ipconfig = connman_ipconfig_create();
+	if (service->ipconfig == NULL) {
+		g_free(service);
+		return NULL;
+	}
+
+	connman_ipconfig_set_method(service->ipconfig,
+					CONNMAN_IPCONFIG_METHOD_DHCP);
+
 	return service;
 }
 
@@ -589,6 +779,13 @@ static gint service_compare(gconstpointer a, gconstpointer b,
 	struct connman_service *service_a = (void *) a;
 	struct connman_service *service_b = (void *) b;
 
+	if (service_a->state != service_b->state) {
+		if (service_a->state == CONNMAN_SERVICE_STATE_READY)
+			return -1;
+		if (service_b->state == CONNMAN_SERVICE_STATE_READY)
+			return 1;
+	}
+
 	if (service_a->order > service_b->order)
 		return -1;
 
@@ -644,6 +841,8 @@ int __connman_service_set_carrier(struct connman_service *service,
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
 	case CONNMAN_SERVICE_TYPE_WIFI:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
 		return -EINVAL;
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 		break;
@@ -666,6 +865,8 @@ int __connman_service_set_carrier(struct connman_service *service,
 int __connman_service_indicate_state(struct connman_service *service,
 					enum connman_service_state state)
 {
+	GSequenceIter *iter;
+
 	DBG("service %p state %d", service, state);
 
 	if (service == NULL)
@@ -692,10 +893,11 @@ int __connman_service_indicate_state(struct connman_service *service,
 
 	if (state == CONNMAN_SERVICE_STATE_READY) {
 		connman_service_set_favorite(service, TRUE);
-		__connman_storage_save_service(service);
 
-		if (service->timeout > 0)
+		if (service->timeout > 0) {
 			g_source_remove(service->timeout);
+			service->timeout = 0;
+		}
 
 		if (service->pending != NULL) {
 			g_dbus_send_reply(connection, service->pending,
@@ -705,11 +907,15 @@ int __connman_service_indicate_state(struct connman_service *service,
 			service->pending = NULL;
 		}
 
+		g_get_current_time(&service->modified);
+		__connman_storage_save_service(service);
 	}
 
 	if (state == CONNMAN_SERVICE_STATE_FAILURE) {
-		if (service->timeout > 0)
+		if (service->timeout > 0) {
 			g_source_remove(service->timeout);
+			service->timeout = 0;
+		}
 
 		if (service->pending != NULL) {
 			DBusMessage *reply;
@@ -721,7 +927,38 @@ int __connman_service_indicate_state(struct connman_service *service,
 			dbus_message_unref(service->pending);
 			service->pending = NULL;
 		}
-	}
+
+		g_get_current_time(&service->modified);
+		__connman_storage_save_service(service);
+	} else
+		service->error = CONNMAN_SERVICE_ERROR_UNKNOWN;
+
+	iter = g_hash_table_lookup(service_hash, service->identifier);
+	if (iter != NULL)
+		g_sequence_sort_changed(iter, service_compare, NULL);
+
+	__connman_profile_changed();
+
+	return 0;
+}
+
+int __connman_service_indicate_error(struct connman_service *service,
+					enum connman_service_error error)
+{
+	DBG("service %p error %d", service, error);
+
+	if (service == NULL)
+		return -EINVAL;
+
+	service->error = error;
+
+	return __connman_service_indicate_state(service,
+					CONNMAN_SERVICE_STATE_FAILURE);
+}
+
+int __connman_service_indicate_default(struct connman_service *service)
+{
+	DBG("service %p", service);
 
 	return 0;
 }
@@ -762,14 +999,12 @@ static struct connman_service *__connman_service_get(const char *identifier)
 		return service;
 	}
 
-	service = g_try_new0(struct connman_service, 1);
+	service = connman_service_create();
 	if (service == NULL)
 		return NULL;
 
 	DBG("service %p", service);
 
-	__connman_service_initialize(service);
-
 	service->identifier = g_strdup(identifier);
 
 	service->profile = g_strdup(__connman_profile_active_ident());
@@ -934,6 +1169,11 @@ struct connman_service *__connman_service_lookup_from_network(struct connman_net
 	return service;
 }
 
+unsigned int __connman_service_get_order(struct connman_service *service)
+{
+	return service->order;
+}
+
 static enum connman_service_type convert_network_type(struct connman_network *network)
 {
 	enum connman_network_type type = connman_network_get_type(network);
@@ -976,8 +1216,8 @@ static enum connman_service_mode convert_wifi_security(const char *security)
 		return CONNMAN_SERVICE_SECURITY_WEP;
 	else if (g_str_equal(security, "wpa") == TRUE)
 		return CONNMAN_SERVICE_SECURITY_WPA;
-	else if (g_str_equal(security, "wpa2") == TRUE)
-		return CONNMAN_SERVICE_SECURITY_WPA2;
+	else if (g_str_equal(security, "rsn") == TRUE)
+		return CONNMAN_SERVICE_SECURITY_RSN;
 	else
 		return CONNMAN_SERVICE_SECURITY_UNKNOWN;
 }
@@ -993,6 +1233,11 @@ static void update_from_network(struct connman_service *service,
 	if (str != NULL) {
 		g_free(service->name);
 		service->name = g_strdup(str);
+		service->hidden = FALSE;
+	} else {
+		g_free(service->name);
+		service->name = NULL;
+		service->hidden = TRUE;
 	}
 
 	service->strength = connman_network_get_uint8(network, "Strength");
@@ -1006,6 +1251,8 @@ static void update_from_network(struct connman_service *service,
 	if (service->strength > strength && service->network != NULL) {
 		connman_network_unref(service->network);
 		service->network = NULL;
+
+		strength_changed(service);
 	}
 
 	if (service->network == NULL) {
@@ -1035,6 +1282,11 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 	const char *ident, *group;
 	char *name;
 
+	if (__connman_service_lookup_from_network(network) != NULL) {
+		connman_error("Service already exists");
+		return NULL;
+	}
+
 	ident = __connman_network_get_ident(network);
 	if (ident == NULL)
 		return NULL;
@@ -1062,6 +1314,8 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 
 	service->type = convert_network_type(network);
 
+	service->state = CONNMAN_SERVICE_STATE_IDLE;
+
 	update_from_network(service, network);
 
 	service_register(service);
@@ -1077,7 +1331,7 @@ static int service_load(struct connman_service *service)
 	GKeyFile *keyfile;
 	gchar *pathname, *data = NULL;
 	gsize length;
-	char *str;
+	gchar *str;
 
 	DBG("service %p", service);
 
@@ -1111,11 +1365,27 @@ static int service_load(struct connman_service *service)
 		break;
 	case CONNMAN_SERVICE_TYPE_WIFI:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
 		service->favorite = g_key_file_get_boolean(keyfile,
 				service->identifier, "Favorite", NULL);
+
+		str = g_key_file_get_string(keyfile,
+				service->identifier, "Failure", NULL);
+		if (str != NULL) {
+			service->state = CONNMAN_SERVICE_STATE_FAILURE;
+			service->error = string2error(str);
+		}
 		break;
 	}
 
+	str = g_key_file_get_string(keyfile,
+				service->identifier, "Modified", NULL);
+	if (str != NULL) {
+		g_time_val_from_iso8601(str, &service->modified);
+		g_free(str);
+	}
+
 	str = g_key_file_get_string(keyfile,
 				service->identifier, "Passphrase", NULL);
 	if (str != NULL) {
@@ -1123,6 +1393,9 @@ static int service_load(struct connman_service *service)
 		service->passphrase = str;
 	}
 
+	__connman_ipconfig_load(service->ipconfig, keyfile,
+					service->identifier, "IPv4.");
+
 	g_key_file_free(keyfile);
 
 	return 0;
@@ -1133,6 +1406,7 @@ static int service_save(struct connman_service *service)
 	GKeyFile *keyfile;
 	gchar *pathname, *data = NULL;
 	gsize length;
+	gchar *str;
 
 	DBG("service %p", service);
 
@@ -1157,21 +1431,48 @@ static int service_save(struct connman_service *service)
 	g_free(data);
 
 update:
+	if (service->name != NULL)
+		g_key_file_set_string(keyfile, service->identifier,
+						"Name", service->name);
+
 	switch (service->type) {
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 		break;
 	case CONNMAN_SERVICE_TYPE_WIFI:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
 		g_key_file_set_boolean(keyfile, service->identifier,
 					"Favorite", service->favorite);
+
+		if (service->state == CONNMAN_SERVICE_STATE_FAILURE) {
+			const char *failure = error2string(service->error);
+			if (failure != NULL)
+				g_key_file_set_string(keyfile,
+							service->identifier,
+							"Failure", failure);
+		} else {
+			g_key_file_remove_key(keyfile, service->identifier,
+							"Failure", NULL);
+		}
 		break;
 	}
 
+	str = g_time_val_to_iso8601(&service->modified);
+	if (str != NULL) {
+		g_key_file_set_string(keyfile, service->identifier,
+							"Modified", str);
+		g_free(str);
+	}
+
 	if (service->passphrase != NULL)
 		g_key_file_set_string(keyfile, service->identifier,
 					"Passphrase", service->passphrase);
 
+	__connman_ipconfig_save(service->ipconfig, keyfile,
+					service->identifier, "IPv4.");
+
 	data = g_key_file_to_data(keyfile, &length, NULL);
 
 	if (g_file_set_contents(pathname, data, length, NULL) == FALSE)
